library flutter_bloc_architecture;

import 'dart:async';

import 'package:flutter/widgets.dart';

abstract class BlocState {
  void dispose() {}
}

abstract class BlocEvent {}

abstract class BlocError {}

abstract class BlocAction<STATE extends BlocState> {
  STATE action(STATE state);

  Bloc<STATE> get bloc => BlocStorageWidget.storage.get(STATE);
}

class Bloc<STATE extends BlocState> {
  STATE _state;
  StreamController<BlocAction> _actionStreamController;
  StreamController<Object> _dataStreamController;

  Bloc(STATE initialState) {
    _actionStreamController = StreamController<BlocAction>.broadcast(sync: false);
    _actionStreamController.stream.listen(_onAction);

    _dataStreamController = StreamController<Object>.broadcast(sync: false);

    this._state = initialState;
  }

  void _onAction(BlocAction action) {
    final state = action.action(this._state);
    if (state != null) {
      this._state = state;
      _dataStreamController.sink.add(this._state);
    }
  }

  void enqueueAction(BlocAction action) {
    if (action != null) {
      _actionStreamController.sink.add(action);
    }
  }

  void enqueueEvent(BlocEvent event) {
    if (event != null) {
      _dataStreamController.sink.add(event);
    }
  }

  void enqueueError(BlocError error) {
    if (error != null) {
      _dataStreamController.sink.addError(error);
    }
  }

  void dispose() {
    if (_actionStreamController != null) {
      _actionStreamController.close();
    }
    if (_dataStreamController != null) {
      _dataStreamController.close();
    }
    if (_state != null) {
      _state.dispose();
    }
  }

  StreamSubscription<Object> listen(void Function(Object data) onData, void Function(Object error) onError) {
    return _dataStreamController.stream.listen(
      onData,
      onError: onError,
      cancelOnError: false,
    );
  }

  STATE get state => _state;
}

class BlocStorage {
  List<Bloc<BlocState>> _list = List();

  void dispose() {
    for (final bloc in _list) {
      bloc.dispose();
    }
    _list.clear();
  }

  Bloc<STATE> create<STATE extends BlocState>(STATE state) {
    final Bloc<STATE> bloc = Bloc<STATE>(state);
    _list.add(bloc);
    return bloc;
  }

  void destroy(Bloc<BlocState> bloc) {
    _list.remove(bloc);
    bloc.dispose();
  }

  Bloc<STATE> get<STATE extends BlocState>(Type type) {
    return _list.firstWhere((bloc) {
      return (bloc.state.runtimeType == type);
    }, orElse: () {
      return null;
    });
  }

  Bloc<STATE> getOrCreate<STATE extends BlocState>(STATE state) {
    return _list.firstWhere((bloc) {
      return (bloc is Bloc<STATE>);
    }, orElse: () {
      final bloc = create(state);
      return bloc;
    });
  }
}

GlobalKey<_BlocStorageWidgetState> _blocStorageAppGlobalKey = GlobalKey();

class BlocStorageWidget extends StatefulWidget {
  final WidgetBuilder builder;

  BlocStorageWidget({this.builder}) : super(key: _blocStorageAppGlobalKey);

  @override
  State<StatefulWidget> createState() {
    return _BlocStorageWidgetState();
  }

  static BlocStorage get storage => _blocStorageAppGlobalKey.currentState.storage;
}

class _BlocStorageWidgetState<STATE extends BlocState> extends State<BlocStorageWidget> {
  BlocStorage storage = BlocStorage();

  @override
  void dispose() {
    storage.dispose();
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    return widget.builder(context);
  }
}

class BlocObserverWidget<STATE extends BlocState> extends StatefulWidget {
  final Bloc<STATE> _bloc;
  final Widget Function(BuildContext context, STATE state) _onBuild;
  final Function(BuildContext context, BlocEvent event, Bloc<BlocState> bloc) _onEvent;
  final Function(BuildContext context, BlocError error, Bloc<BlocState> bloc) _onError;

  BlocObserverWidget({
    @required Bloc<STATE> bloc,
    @required Widget Function(BuildContext context, STATE state) onBuild,
    Function(BuildContext context, BlocEvent event, Bloc<BlocState> bloc) onEvent,
    Function(BuildContext context, BlocError error, Bloc<BlocState> bloc) onError,
  })  : _bloc = bloc,
        _onBuild = onBuild,
        _onEvent = onEvent,
        _onError = onError;

  @override
  State<StatefulWidget> createState() {
    return _BlocObserverWidgetState<STATE>(_onBuild, _onEvent, _onError);
  }
}

class _BlocObserverWidgetState<STATE extends BlocState> extends State<BlocObserverWidget> {
  StreamSubscription<Object> _streamSubscription;
  final Widget Function(BuildContext context, STATE state) _onBuild;
  final Function(BuildContext context, BlocEvent event, Bloc<BlocState> bloc) _onEvent;
  final Function(BuildContext context, BlocError error, Bloc<BlocState> bloc) _onError;

  _BlocObserverWidgetState(this._onBuild, this._onEvent, this._onError);

  @override
  void initState() {
    super.initState();
    _streamSubscription = widget._bloc.listen(_onStreamData, _onStreamError);
  }

  @override
  void dispose() {
    _streamSubscription.cancel();
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    return _onBuild(context, widget._bloc.state);
  }

  void _onStreamData(Object data) {
    if (data is BlocState) {
      setState(() {});
    } else if (data is BlocEvent) {
      if (_onEvent != null) {
        _onEvent(context, data, widget._bloc);
      }
    }
  }

  void _onStreamError(Object error) {
    _onError(context, error, widget._bloc);
  }
}
